""" Utility functions used by PXTL implementations.
"""

__all__= [
  'attrList', 'removeReparent', 'setDoctype', 'getNamespaces',
  'DictAsInstance', 'encodeText', 'decodeName', 'parsePseudoPIs',
  'ConditionalContext', 'checkFeatures', 'haveNestedScopes',
  'AbstractPXTLObject', 'determineMethod', 'writeTypeHeader',
  'getLogicalLines', 'getIndentSize', 'renumberCode', 'evalStatic', 'lastItem'
]

from pxtl.constants import *
from pxtl.exceptions import *

import __future__
import types, sys, string, re, urllib, urlparse, pxtl, new


# DOM utility functions
#
def attrList(nodes):
  """ Turn a live DOM NamedNodeMap into a static Python list, which can be
      used for potentially destructive iteration.
  """
  l= []
  for i in range(nodes.length):
    l.append(nodes.item(i))
  return l


def removeReparent(node):
  """ Replace a DOM node with its children, making sure all namespaces are
      the same on the child by copying declarations where necessary.
  """
  for i in range(node.attributes.length):
    attribute= node.attributes.item(i)
    if attribute.namespaceURI==NSNS and attribute.value!=PXNS:

      # Ignore declarations that are already present on parent
      #
      if node.parentNode is not None:
        if node.parentNode.nodeType!=node.DOCUMENT_NODE:
          if attribute.prefix is None:
            prefix= None
          else:
            prefix= attribute.localName
          value= attribute.value
          if value=='':
            value= None
          if node.parentNode.lookupNamespaceURI(prefix)==value:
            continue

      # Ignore declarations that are overridden (declared on child). Copy the
      # rest to each child.
      #
      for child in list(node.childNodes):
        if child.nodeType==child.ELEMENT_NODE:
          if not child.hasAttributeNS(NSNS, attribute.localName):
            child.setAttributeNS(NSNS, attribute.name, attribute.value)

  # Replace node with a DocumentFragment of all its children.
  #
  parent= node.parentNode
  frag= node.ownerDocument.createDocumentFragment()
  for child in list(node.childNodes):
    if parent.nodeType!=node.DOCUMENT_NODE or child.nodeType!=node.TEXT_NODE:
      frag.appendChild(child)
  parent.replaceChild(frag, node)


def setDoctype(document, doctype):
  """ Try to set the doctype of a document to the specified object. This can't
      be done directly in DOM Level 2, and in Level 3 it's still not required,
      so direct replacement may not be possible with implementations other
      than pxdom. In this case, clone the document's content into a new
      document with the correct doctype.
  """
  current= document.doctype
  try:
    if current is None:
      if doctype is None:
        return document
      else:
        document.insertBefore(doctype, document.firstChild)
    else:
      if doctype is None:
        document.removeChild(current)
      else:
        document.replaceChild(doctype, current)
    return document
  except NonErrors:
    raise
  except Exception:

    # Failed in some way (why can't tell what, since DOMExceptions are not
    # bound to anything in the Python bindings - the exception could be
    # anything). Try the copying method instead.
    #
    imp= document.implementation
    newDocument= imp.createDocument(PXNS, 'px:none', doctype)
    newDocument.xmlVersion= document.xmlVersion
    refEl= newDocument.documentElement
    for child in list(document.childNodes):
      if child.nodeType==child.ELEMENT_NODE:
        newDocument.replaceChild(
          newDocument.importNode(child, True), newDocument.documentElement
        )
        refEl= None
      elif child.nodeType!=child.DOCUMENT_TYPE_NODE:
        newDocument.insertBefore(newDocument.importNode(child, True), refEl)
    return newDocument


def getNamespaces(node):
  """ Get a mapping {prefix: uri} of all non-built-in namespaces active on an
      element.
  """
  nss= {}
  while node.nodeType==node.ELEMENT_NODE:
    for attr in attrList(node.attributes):
      if attr.namespaceURI==NSNS:
        prefix= attr.localName
        if attr.localName=='xmlns':
          prefix= None
        if not nss.has_key(prefix):
          nss[prefix]= attr.value
    node= node.parentNode
  return nss


# Text encoding
#
def encodeText(text, coding= 'text'):
  """ Return text in an escaped form (for URL parameters, JS literals etc.).
  """
  if coding=='text':
    return text
  if coding in ('upar', 'name') and isinstance(text, unicode):
    text= text.encode('utf-8')
  try:
    find, sub= _CODINGS[coding]
    text= find.sub(sub, text)
  except KeyError:
    raise ValueError ("unknown coding '%s'" % coding)
  if coding=='name':
    if text=='':
      return '___'
    if text[0] in '0123456789-.':
      text= ('_%02X' % ord(text[0]))+text[1:]
  return text

def decodeName(text):
  if isinstance(text, unicode):
    text= text.encode('utf-8', 'ignore')
  text= _decode_name.sub(_decode_name_s, text)
  return unicode(text, 'utf-8', 'ignore')


_hexdigits= '0123456789ABCDEF0123456789abcdef'

_encode_upar= re.compile(r'[^a-zA-Z0-9\-\._]')
def _encode_upar_s(m):
  return '%%%02X' % ord(m.group(0))

_encode_cstr= re.compile(r'[^ !#$%\x28-\x2C\x2E-\x3B\x3D\x3F-\x5B\x5D-\x7E]')
def _encode_cstr_s(m):
  c= ord(m.group(0))
  if c<0x100:
    return r'\%02X ' % c
  elif c<0x10000:
    return r'\%04X ' % c
  else: # will only work when Python is wide-build
    return r'\%06X' % c

_encode_jstr= _encode_cstr
def _encode_jstr_s(m):
  c= ord(m.group(0))
  if c<0x100:
    return r'\x%02X' % c
  elif c<0x10000:
    return r'\u%04X' % c
  else: # JS has no wide unicode support - output surrogates
    c= c-0x10000
    return r'\u%04X\u%04X' % (0xD800+(c >> 10), 0xDC00+(c % 0x400))

_encode_name= re.compile(r'[^a-zA-Z0-9\-\.]')
def _encode_name_s(m):
  return '_%02X' % ord(m.group(0))

_decode_name= re.compile(r'_..')
def _decode_name_s(m):
  c= m.group(0)[1:]
  if c[0] in _hexdigits and c[1] in _hexdigits:
    return chr(eval('0x'+c))
  return ''

_CODINGS= {
  'upar': (_encode_upar, _encode_upar_s),
  'name': (_encode_name, _encode_name_s),
  'cstr': (_encode_cstr, _encode_cstr_s),
  'jstr': (_encode_jstr, _encode_jstr_s),
}

# General-purpose utility functions.
#
class DictAsInstance:
  def __init__(self, d):
    self.__dict__= d

def mimeParameters(header):
  """ Read the main-value and parameter dictionary from a parameterised header

  Provided here because the mimetools implementation is not easily accessible
  outside of Message objects, and to cope with some of the non-standard rubbish
  encountered in the wild. Forget RFCs 2388, 2047 and 2231...
  """
  parts= string.split(header, ';', 1)
  main= string.lower(string.strip(parts[0]))
  parlookup= {}
  if len(parts)==2:
    pars= parts[1]
    ix= 0
    while ix<len(pars):
      ix0= pars.find('=', ix)
      if ix0==-1:
        break
      name= pars[ix:ix0].strip()

      # Walk over parameters list, looking for '";' to end a parameter if
      # possible, to reduce the "-in-string problem
      #
      ix0= ix0+1
      if pars[ix0:ix0+1]!='"':
        ix1= pars.find(';', ix0)
        if ix1==-1:
          ix=ix1= len(pars)
        else:
          ix= ix1+1
      else:
        ix0= ix0+1
        ix1= pars.find('";', ix0)
        if ix1!=-1:
          ix= ix1+2
        else:
          ix1= pars.find('"', ix0)
          if ix1==-1:
            ix=ix1= len(pars)
          else:
            ix= pars.find(';', ix1+1)
            if ix==-1:
              ix= len(pars)
      value= pars[ix0:ix1]

      if name!='':
        parlookup[name.lower()]= value
  return main, parlookup


_ws= [' ', '\n', '\t']
def parsePseudoPIs(v, node= None):
  """ Break an attribute value up into text and pseudo-PI sections
      (target, data).
  """
  parts= []
  ix= 0
  while True:
    ox= string.find(v, '{?', ix)
    if ox==-1:
      if ix<len(v):
        parts.append(v[ix:])
      break
    if ox>ix:
      parts.append(v[ix:ox])
    cx= string.find(v, '?}', ox+1)
    if cx==-1:
      raise PXTLPseudoPIOpenError(node, ox)
    if cx==ox+1:
      parts.append('{?')
    else:
      wx= ox+2
      while wx<cx:
        if v[wx] in _ws:
          break
        wx= wx+1
      vx= wx
      while vx<cx:
        if v[vx] not in _ws:
          break
        vx= vx+1
      parts.append((v[ox+2:wx], v[vx:cx]))
    ix= cx+2
  return parts


def checkFeatures(future):
  """ Convert a future attribute node into a flags word as used by compile().

  Raise an exception for any features that cannot be used in the current
  version of Python.
  """
  flags= 0
  if future==None:
    return flags
  features= map(string.strip, string.split(future.value, ','))
  for feature in features[:]:

    # If the feature is not implemented at all in the current version of
    # Python, complain. If the feature is on by default anyway, don't bother
    # including it in the flags.
    #
    if hasattr(__future__, feature):
      f= getattr(__future__, feature)
      if sys.version_info<f.getOptionalRelease():
        raise PXTLFutureFeatureError(future, feature)
      if sys.version_info<f.getMandatoryRelease():
        flags|= f.compiler_flag
  return flags

def haveNestedScopes(flags):
  """ Determine whether PXTL should provide nested scopes to subtemplates

  Check the supplied flags word and the current Python version's default
  features.
  """
  f= __future__.nested_scopes
  if sys.version_info>=f.getMandatoryRelease():
      return True
  return (flags & f.compiler_flag)!=0


class ConditionalContext:
  """ Logic for if-orif-anif-elif-else extended conditionals.

  Keeps track of previous clauses and calculates the success of each element.
  Used by conditional elements and conditional attributes.
  """
  def reset(self):
    """ Start a new conditional context, forgetting any previous clauses.

    This also serves as the constructor.
    """
    self.old_cond= None
    self.old_testing= None
    self.old_success= None
  __init__= reset

  def getTesting(self, cond, node):
    """ Decide wheth the test condition even needs to be evaluated.

    Given the new clause 'cond' and knowledge of previous clauses, it may not
    actually need to be tested to determine its success. If not, a laziness
    short-circuit kicks in and we can avoid it.
    """
    if cond=='if':
      return True
    if self.old_cond not in PRECONDITIONALS:
      raise PXTLConditionalOrderError(node, self.old_cond)
    if cond=='else':
      return False
    if cond=='orif':
      return not self.old_success
    if cond=='anif':
      return self.old_success
    if cond=='elif':
      if self.old_cond=='elif':
        return self.old_testing and not self.old_success
      else:
        return not self.old_success
    raise ValueError('Unknown conditional %s' % cond)

  def getSuccess(self, cond, test):
    """ Work out whether a clause is successful

    Given the truth of its evaluated test condition. If the condition was not
    evaluated (due to getTesting having returned False), 'test' should be
    None instead of a boolean. Store the clause info as it may affect a
    following clause.
    """
    if cond=='else':
      if self.old_cond=='elif':
        success= self.old_testing and not self.old_success
      else:
        success= not self.old_success
    elif cond=='orif':
      success= test in (None, True)
    else:
      success= test==True
    self.old_cond= cond
    self.old_testing= test is not None
    self.old_success= success
    return success

  def copy(self):
    """ Make a copy of a ConditionalContext

    This can be further updated without affecting this one (for nesting).
    """
    clone= ConditionalContext()
    clone.old_cond= self.old_cond
    clone.old_testing= self.old_testing
    clone.old_success= self.old_success
    return clone


class AbstractPXTLObject:
  """ Object passed into all template global scopes

  For output-writing and constant-holding purposes.
  """
  version= 1,1
  pxtl= pxtl

  TEXT=    ('text/plain',            'text', None, None)
  CSS=     ('text/css',              'text', None, None)
  JS=      ('text/javascript',       'text', None, None)
  XML=     ('text/xml',              'xml',  None, None)

  HTML2=   ('text/html',             'html', '-//IETF//DTD HTML 2.0//EN', None)
  HTML3=   ('text/html',             'html', '-//W3C//DTD HTML 3.2 Final//EN', None)
  TAGSOUP= ('text/html',             'html', '-//W3C//DTD HTML 4.01 Transitional//EN', None)

  HTML4S=  ('text/html',             'html', '-//W3C//DTD HTML 4.01//EN', 'http://www.w3.org/TR/html4/strict.dtd')
  HTML4T=  ('text/html',             'html', '-//W3C//DTD HTML 4.01 Transitional//EN', 'http://www.w3.org/TR/html4/loose.dtd')
  HTML4F=  ('text/html',             'html', '-//W3C//DTD HTML 4.01 Frameset//EN', 'http://www.w3.org/TR/html4/frameset.dtd')
  HTML5=   ('text/html',             'html', False, None)

  XHTML1S= ('text/html',             'xhtml','-//W3C//DTD XHTML 1.0 Strict//EN', 'http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd')
  XHTML1T= ('text/html',             'xhtml','-//W3C//DTD XHTML 1.0 Transitional//EN', 'http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd')
  XHTML1F= ('text/html',             'xhtml','-//W3C//DTD XHTML 1.0 Frameset//EN', 'http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd')
  XHTML11= ('text/html',             'xhtml','-//W3C//DTD XHTML 1.1//EN', 'http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd')
  XHTML1R= ('text/html',             'xhtml','-//W3C//DTD XHTML+RDFa 1.0//EN', 'http://www.w3.org/MarkUp/DTD/xhtml-rdfa-1.dtd')
  XHTML5=  ('text/html',             'xhtml',False, None)

  XXHTML1S=('application/xhtml+xml', 'xml',  '-//W3C//DTD XHTML 1.0 Strict//EN', 'http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd')
  XXHTML1T=('application/xhtml+xml', 'xml',  '-//W3C//DTD XHTML 1.0 Transitional//EN', 'http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd')
  XXHTML1F=('application/xhtml+xml', 'xml',  '-//W3C//DTD XHTML 1.0 Frameset//EN', 'http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd')
  XXHTML11=('application/xhtml+xml', 'xml',  '-//W3C//DTD XHTML 1.1//EN', 'http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd')
  XXHTML11=('application/xhtml+xml', 'xml',  '-//W3C//DTD XHTML 1.1//EN', 'http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd')
  XXHTML1R=('application/xhtml+xml', 'xml',  '-//W3C//DTD XHTML+RDFa 1.0//EN', 'http://www.w3.org/MarkUp/DTD/xhtml-rdfa-1.dtd')
  XXHTML5= ('application/xhtml+xml', 'xml',  False, None)

  XHTMLB1= ('application/xhtml+xml', 'xhtml','-//W3C//DTD XHTML Basic 1.0//EN', 'http://www.w3.org/TR/xhtml-basic/xhtml-basic10.dtd')
  XHTMLM1= ('application/xhtml+xml', 'xhtml','-//WAPFORUM//DTD XHTML Mobile 1.0//EN', 'http://www.wapforum.org/DTD/xhtml-mobile10.dtd')
  XHTMLM12=('application/xhtml+xml', 'xhtml','-//WAPFORUM//DTD XHTML Mobile 1.2//EN', 'http://www.openmobilealliance.org/tech/DTD/xhtml-mobile12.dtd')
  IXHTML1= ('application/xhtml+xml', 'xhtml','-//i-mode group (ja)//DTD XHTML i-XHTML (Locale/Ver.=ja/1.0) 1.0//EN', 'i-xhtml_4ja_10.dtd')
  IXHTML11=('application/xhtml+xml', 'xhtml','-//i-mode group (ja)//DTD XHTML i-XHTML (Locale/Ver.=ja/1.1) 1.0//EN', 'i-xhtml_4ja_10.dtd')
  WML11=   ('application/xhtml+xml', 'xml',  '-//WAPFORUM//DTD WML 1.1//EN', 'http://www.wapforum.org/DTD/wml_1.1.xml')
  WML12=   ('application/xhtml+xml', 'xml',  '-//WAPFORUM//DTD WML 1.2//EN', 'http://www.wapforum.org/DTD/wml12.dtd')
  WML13=   ('application/xhtml+xml', 'xml',  '-//WAPFORUM//DTD WML 1.3//EN', 'http://www.wapforum.org/DTD/wml13.dtd')
  WML20=   ('application/xhtml+xml', 'xml',  '-//WAPFORUM//DTD WML 2.0//EN', 'http://www.wapforum.org/dtd/wml20.dtd')

  MATH1=   ('application/mathml+xml','xml',  None, 'http://www.w3.org/Math/DTD/mathml1/mathml.dtd')
  MATH2=   ('application/mathml+xml','xml',  '-//W3C//DTD MathML 2.0//EN', 'http://www.w3.org/TR/MathML2/dtd/mathml2.dtd')
  SVG1=    ('image/svg+xml',         'xml',  '-//W3C//DTD SVG 1.0//EN', 'http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd')
  SVG11=   ('image/svg+xml',         'xml',  '-//W3C//DTD SVG 1.1//EN', 'http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd')
  SVG11B=  ('image/svg+xml',         'xml',  '-//W3C//DTD SVG 1.1 Basic//EN', 'http://www.w3.org/Graphics/SVG/1.1/DTD/svg11-basic.dtd')
  SVG11T=  ('image/svg+xml',         'xml',  '-//W3C//DTD SVG 1.1 Tiny//EN', 'http://www.w3.org/Graphics/SVG/1.1/DTD/svg11-tiny.dtd')
  XHMS=    ('application/xhtml+xml', 'xml',  '-//W3C//DTD XHTML 1.1 plus MathML 2.0 plus SVG 1.1//EN', 'http://www.w3.org/2002/04/xhtml-math-svg/xhtml-math-svg.dtd')



def determineMethod(doctuple, namespaceURI, localName):
  """ Choose which output method to use

  Given an evaluated doctype attribute and the name of the root node.
  """
  method= doctuple[1]
  if method is not None:
    return method
  if namespaceURI==HTNS or (namespaceURI is None and localName=='html'):
    return 'xhtml'
  else:
    return 'xml'
  
  
def writeTypeHeader(writer, mimetype, method, encoding):
  """ Send an HTTP Content-Type response header best matching the document.
  """
  if mimetype is None:
    if method in ('html', 'xhtml'):
      mimetype= 'text/html'
    elif method=='xml':
      mimetype= 'text/xml'
    else:
      mimetype= 'text/plain'
  if encoding is not None and string.lower(mimetype[:5])=='text/':
    writer.write(
      'Content-Type: %s; charset=%s\r\n\r\n' % (str(mimetype), str(encoding))
    )
  else:
    writer.write('Content-Type: %s\r\n\r\n' % str(mimetype))


def getLogicalLines(block, flags= 0):
  """ Split a multi-line code block string into Python logical lines

  Removes any common indentation.
  """
  # (Non-logical lines include empty lines, comment lines, triple-quoted
  # strings, backslash-continuations and bracketed expressions.)
  #
  # Finding out which lines are logical lines in Python code is not normally
  # possible. We cheat by compiling the code over and over again, each time
  # changing the spacing of one line. If that line is a logical line, changing
  # the spacing will cause an IndentationError. If no error occurs we know
  # that physical line is not a logical line.
  #
  # This can be a bit slow, so it is only used in the optimised implementation
  # at compile-time. The reference implementation can get away with a
  # simpler hack than redentation as its code block indent levels are flat.

  # Get the common indentation by looking at the indentation of the first
  # non-empty, non-comment line.
  #
  lines= string.split(block, '\n')
  for line in lines:
    undent= getIndentSize(line)
    if undent is not None:
      break
  else:
    return [block]

  # Make an indented copy of the lines so they can definitely be included in
  # an if-block. (Use eight spaces so that there can be no interaction with
  # any following tabs). Check the results are compilable before we start
  # messing with the indentation.
  #
  execblock= 'if 1:\n        %sdoNothing\n%%s\n' % (' '*undent)
  testlines= map(lambda x: '        '+x, lines)
  compile(execblock % string.join(testlines, '\n'), '<test>', 'exec', flags)

  # Iterate over lines of code, screwing up a single line's indentation at a
  # time by replacing any indentation it has with a single space character;
  # this can never match a consistent indent level (as we have indented the
  # rest by 8 spaces).
  #
  logicals= []
  sink= StreamSink()
  for ix in range(len(testlines)):
    line= testlines[ix]
    testlines[ix]= ' '+string.lstrip(line)

    # Try to compile the result. If there's an error, add the current line to
    # the list of logical lines, first removing any common indentation.
    # Aggravatingly, Python 1.6 and earlier send the message 'inconsistent
    # dedent' to stderr as well as throwing the exception; we don't want this,
    # so temporarily redirect stderr to oblivion.
    #
    try:
      if sys.hexversion<0x02000000:
        oldStderr= sys.stderr
        sys.stderr= sink
      try:
        compile(execblock % string.join(testlines, '\n'), '<test>', 'exec', flags)
      finally:
        if sys.hexversion<0x02000000:
          sys.stderr= oldStderr
    except SyntaxError:
      logicals.append(
        ' '*(getIndentSize(lines[ix])-undent) + string.lstrip(lines[ix])
      )

    # If no error, add the physical line to the previous logical line (if any)
    #
    else:
      if logicals==[]:
        logicals.append(lines[ix])
      else:
        logicals[-1]= logicals[-1]+'\n'+lines[ix]

    # Restore the indentation and move onto the next physical line.
    #
    testlines[ix]= line
  return logicals


def getIndentSize(line):
  """ Return the number of spaces a line's indent is equivalent to. If the
      line is empty or a comment line, the indentation is irrelevant and None
      is returned.
  """
  stripped= string.lstrip(line)
  if stripped=='' or stripped[:1]=='#':
    return None
  indent= line[:len(line)-len(stripped)]
  if '\t' not in indent:
    return len(indent)
  size= 0
  for c in indent:
    if c=='\t':
      size= size+TABSIZE-(size%TABSIZE)
    else:
      size= size+1
  return size
TABSIZE= 8 # currently defined by Python


class StreamSink:
  """ A /dev/null stream as a class.
  """
  def write(self, s):
    pass
  def read(self):
    return ''


def evalStatic(expr, flags= 0):
  """ Compile and evaluate an expression if it is completely static, that is
      it does not depend on or alter anything from a scope. This means any
      expression containing only literals. Expressions that cause errors are
      never static. Return a tuple of is-static-flag and static value.
  """

  # Compile the expression. If it contains variable names or attribute access
  # it will store names used
  #
  try:
    code= compile(string.strip(expr), '<evalStatic>', 'eval', flags)
  except SyntaxError:
    return (False, None)

  # Look at main code object and any sneaky lambdas hidden in it
  #
  codes= [code]
  for const in code.co_consts:
    if type(const)==type(code):
      codes.append(const)

  # Search code for use of names. For the benefit of the common case, pretend
  # that None, False and True are literals, not values. (As they may be in
  # future Pythons.)
  #
  for c in codes:
    for name in c.co_names:
      if name not in ('None', 'False', 'True'):
        return (False, None)

  # OK, looks static. Let's try to execute it. If it errors now let's say it
  # isn't static, so the error can be generated at the expected run-time
  # instead.
  #
  try:
    return (True, eval(code, STATICSCOPE))
  except NonErrors:
    raise
  except Exception:
    return (False, None)
STATICSCOPE= {'None': None, 'False': False, 'True': True}


def renumberCode(code, linetrans):
  """ Change the line numbers set in a Python code object

  Pass the lines of the object (and any sub-code object it contains) through a
  translation list so they match the positions in the original PXTL template
  instead of the temporary code block they were compiled with. This allows
  execution frames (including exception tracebacks) to give useful line
  numbers.
  """
  lnotab= StringIO()

  progline= code.co_firstlineno
  if len(linetrans)<=progline:
    progline= len(linetrans)-1
  firstlineno= linetrans[progline]

  srcline= firstlineno
  ix= 0
  while ix<len(code.co_lnotab)-1:
    bytes, lines= ord(code.co_lnotab[ix]), ord(code.co_lnotab[ix+1])
    ix= ix+2

    progline= progline+lines
    if len(linetrans)<=progline:
      progline= len(linetrans)-1
    increment= linetrans[progline]-srcline
    if increment<0:
      increment= 0
    srcline= linetrans[progline]

    while increment>=256:
      lnotab.write(chr(bytes))
      lnotab.write('\xFF')
      increment= increment-255
      bytes= 0
    lnotab.write(chr(bytes))
    lnotab.write(chr(increment))

  # Check for other constant code blocks (def statements) compiled into this
  # one and fix them up too.
  #
  consts= []
  for const in code.co_consts:
    if type(const)==type(code):
      const= renumberCode(const, linetrans)
    consts.append(const)

  # Rewrap the code object with altered line numbers and constants. Remember
  # to include cellvars and freevars if they are available (Python 2.1+); this
  # will need to be extended if the code object's properties are extended in a
  # future version of Python.
  #
  args= [
    code.co_argcount, code.co_nlocals, code.co_stacksize, code.co_flags,
    code.co_code, tuple(consts), code.co_names, code.co_varnames,
    code.co_filename, code.co_name, firstlineno, lnotab.getvalue()
  ]
  if hasattr(code, 'co_freevars'):
    args.append(code.co_freevars)
  if hasattr(code, 'co_cellvars'):
    args.append(code.co_cellvars)
  return apply(new.code, args)


# Utility function to return the last non-None item in a list, used for run-
# time whitespace handling.
#
def lastItem(l):
  for i in range(len(l)):
    if l[len(l)-i-1] is not None:
      return l[len(l)-i-1]
  return None

